#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from enum import Enum, auto
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
from numpy.typing import DTypeLike, NDArray


class CamVariant(Enum):
    """All supported variants of CAMs."""

    ACAM = auto()
    """Analog CAM."""

    TCAM = auto()
    """Ternary CAM."""

    @property
    def cell_encoding_width(self) -> int:
        """
        The amount of elements in a CAM row that is used
        to encode a single CAM cell.
        """

        if self == CamVariant.ACAM:
            return 2

        return 1


class CamOp(Enum):
    """All supported operations on CAMs."""

    MATCH = auto()
    """Get the match value for each row."""

    COUNT_MISMATCHES = auto()
    """
    Count the number of mismatches per row.
    For TCAMs, this is also called the hamming distance.
    """

    REDUCE_SUM = auto()
    """
    Get the match value per row and further reduce each input
    down to a single value, picked from a vector of values
    adjacent to the CAM.
    """

    @property
    def is_reduction(self) -> bool:
        """Whether the op is a reduction."""
        return self == CamOp.REDUCE_SUM

    def result_dtype(self, reduction_values: Optional[NDArray] = None) -> DTypeLike:
        """
        The resulting dtype of the op.
        For a reduction, this depends on the values that are reduced over.
        """

        if self == CamOp.MATCH:
            return np.int8

        elif self == CamOp.COUNT_MISMATCHES:
            return np.int64

        elif self == CamOp.REDUCE_SUM:
            if reduction_values is None:
                raise ValueError(
                    "could not determine resulting data type for CAM operation"
                )
            return reduction_values.dtype


FloatDType = Union[np.float32, np.float64]
"""All float NumPy data types."""

IntDType = Union[
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]
"""All integer NumPy data types."""

NumericDType = Union[IntDType, FloatDType]
"""All numeric NumPy data types, i.e., floats and ints."""

Dimensions = Tuple[int, int, int]
"""A 3-dimensional vector used for CUDA launch configurations."""

LaunchConfiguration = Tuple[Dimensions, Dimensions]
"""A CUDA kernel launch configuration in `(dim_grid, dim_block)` format."""

Kernel = Callable[[Dimensions, Dimensions, Tuple[Any, ...]], None]
"""A `cupy.RawKernel` instance."""

DTYPE_TO_CTYPE = {
    np.float32: "float",
    np.float64: "double",
    np.int8: "char",
    np.int16: "short",
    np.int32: "int",
    np.int64: "long",
    np.uint8: "char",
    np.uint16: "unsigned short",
    np.uint32: "unsigned int",
    np.uint64: "unsigned long",
}
"""Lookup table for numpy dtypes -> their type name in C."""


def dtype_to_ctype(dtype: DTypeLike) -> str:
    """Converts a numpy dtype to the name of its counterpart in C."""
    # directly indexing `_DTYPE_TO_CTYPE` doesn't work with dtype instances,
    # such as those obtained via `arr.dtype`. They need to be directly compared.
    for nptype in DTYPE_TO_CTYPE:
        if dtype == nptype:
            return DTYPE_TO_CTYPE[nptype]

    raise TypeError(f"data type {dtype} is not supported")


def is_float_type(dtype: DTypeLike) -> bool:
    """Determines whether a given data type is a floating point type."""
    return dtype == np.float32 or dtype == np.float64



#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""Helper functionality."""

from math import ceil

from ..types import LaunchConfiguration


def simple_kernel_dimensions(
    threads_needed: int, max_threads_per_block: int
) -> LaunchConfiguration:
    """
    Determines a simple one-dimensional set of CUDA kernel dimensions,
    i.e. `(dim_grid, dim_block)` based on a wanted amount of threads
    """

    if threads_needed <= 0:
        return ((0, 0, 0), (0, 0, 0))

    threads_per_block = min(threads_needed, max_threads_per_block)
    blocks_per_grid = ceil(threads_needed / threads_per_block)

    dim_grid = (blocks_per_grid, 1, 1)
    dim_block = (threads_per_block, 1, 1)

    return dim_grid, dim_block



#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""Code generation for CUDA kernels that handle CAM operations."""

from string import Template
from typing import Dict, Tuple

import cupy as cp
from numpy.typing import DTypeLike

from .types import CamOp, CamVariant, Kernel, dtype_to_ctype, is_float_type

# Kernel notes:
#
# Dimensions:
#     The indexing inside of a two-dimensional pair of an input matrix and a CAM
#     is done linearly via `threadIdx.x` and `blockIdx.x`.
#     The indexing within the (flattened) stack of input and CAM pairs is done
#     via `blockIdx.y`.
#
# Operation per thread:
#     Each thread reduces an input/CAM column down to a single match.
#
# Shapes:
#     The kernel by itself only matches inputs and CAMs of the same dimensions,
#     including dimensions larger than two. To use this kernel with uneven dimensions,
#     it needs to be wrapped in additional code that flattens the uneven arguments
#     down to even ones and then reshapes the result back into the demanded shape.

KERNEL_OUTLINE = Template(
    r"""
extern "C" __global__
void cam(
    $inputs_type *inputs, $cam_type *cam,
    long columns, long input_rows, long cam_rows
    $extra_params
) {
    long block_id = blockIdx.x + blockIdx.y * gridDim.x;
    long thread_id = block_id * (blockDim.x * blockDim.y) + threadIdx.x;

    /* `gridDim.y` encodes the amount of cores in the stack.
        Return here in case the thread is not needed. */
    if (thread_id >= input_rows * cam_rows * gridDim.y) {
        return;
    }

    /* these are absolute indices, i.e., within the entire stack of inputs */
    long cam_row_index = blockIdx.y * cam_rows + thread_id % cam_rows;
    long input_row_index = thread_id / cam_rows;

    $pre_loop

    for (long i = 0; i < columns; i++) {
        $loop_contents
    }

    $post_loop
}
"""
)

TCAM_INT_MATCHING = Template(
    r"""
        $cam_type cam_value = cam[cam_row_index * columns + i];
        $inputs_type input_value = inputs[input_row_index * columns + i];

        if (cam_value > 1 || cam_value < 0) {
            continue;
        }

        if (input_value - cam_value != 0) {
            $on_mismatch
        }
"""
)

TCAM_FLOAT_MATCHING = Template(
    r"""
        $cam_type cam_value = cam[cam_row_index * columns + i];
        $inputs_type input_value = inputs[input_row_index * columns + i];

        if (fabs(input_value - cam_value) > 0.00001) {
            $on_mismatch
        }
"""
)

ACAM_FLOAT_MATCHING = Template(
    r"""
        $cam_type min_threshold = cam[2 * (cam_row_index * columns + i)];
        $cam_type max_threshold = cam[2 * (cam_row_index * columns + i) + 1];
        $inputs_type input_value = inputs[input_row_index * columns + i];

        /* nan = don't care */
        if (!(
            (isnan(min_threshold) || min_threshold <= input_value) &&
            (isnan(max_threshold) || input_value < max_threshold)
        )) {
            $on_mismatch
        }
"""
)

ACAM_INT_MATCHING = Template(
    r"""
        $cam_type min_threshold = cam[2 * (cam_row_index * columns + i)];
        $cam_type max_threshold = cam[2 * (cam_row_index * columns + i) + 1];
        $inputs_type input_value = inputs[input_row_index * columns + i];

        /* < 0 = don't care */
        if (!(
            (min_threshold < 0 || min_threshold <= input_value) &&
            (max_threshold < 0 || input_value < max_threshold)
        )) {
            $on_mismatch
        }
"""
)


def generate_kernel(
    variant: CamVariant,
    op: CamOp,
    inputs_dtype: DTypeLike,
    cam_dtype: DTypeLike,
    results_dtype: DTypeLike,
) -> Kernel:
    """Generates CUDA kernel code and returns a callable CAM kernel."""

    inputs_type = dtype_to_ctype(inputs_dtype)
    cam_type = dtype_to_ctype(cam_dtype)
    results_type = dtype_to_ctype(results_dtype)

    MATCHING_KINDS: Dict[Tuple[CamVariant, bool], Template] = {
        (CamVariant.TCAM, True): TCAM_FLOAT_MATCHING,
        (CamVariant.TCAM, False): TCAM_INT_MATCHING,
        (CamVariant.ACAM, True): ACAM_FLOAT_MATCHING,
        (CamVariant.ACAM, False): ACAM_INT_MATCHING,
    }

    matching = MATCHING_KINDS[(variant, is_float_type(cam_dtype))]

    if op == CamOp.MATCH:
        extra_params = f", {results_type} *matches"
        pre_loop = f"{results_type} match = 1;"
        post_loop = "matches[thread_id] = match;"
        on_mismatch = "match = 0; break;"

    elif op == CamOp.COUNT_MISMATCHES:
        extra_params = f", {results_type} *counts"
        pre_loop = f"{results_type} count = 0;"
        post_loop = "counts[thread_id] = count;"
        on_mismatch = "count++;"

    elif op == CamOp.REDUCE_SUM:
        extra_params = f", {results_type} *results, {results_type} *values"
        pre_loop = ""
        post_loop = "atomicAdd(&results[input_row_index], values[cam_row_index]);"
        on_mismatch = "return;"

    else:
        raise TypeError(f"unknown CAM operation: {op}")

    loop_contents = matching.safe_substitute(
        inputs_type=inputs_type,
        cam_type=cam_type,
        on_mismatch=on_mismatch,
    )

    code = KERNEL_OUTLINE.safe_substitute(
        inputs_type=inputs_type,
        cam_type=cam_type,
        results_type=results_type,
        extra_params=extra_params,
        loop_contents=loop_contents,
        pre_loop=pre_loop,
        post_loop=post_loop,
    )

    return cp.RawKernel(code=code, name="cam")



#
#  Copyright (2024) Hewlett Packard Enterprise Development LP
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  You may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""Python wrapper around CAM kernels."""

from math import ceil, prod
from typing import Optional, Tuple

import cupy as cp
from numpy.typing import DTypeLike, NDArray

from .types import CamOp, CamVariant, Kernel


def run_kernel(
    kernel: Kernel,
    variant: CamVariant,
    op: CamOp,
    inputs: NDArray,
    cam: NDArray,
    result_dtype: DTypeLike,
    reduction_values: Optional[NDArray] = None,
) -> NDArray:
    """
    This is the heart of this library's shape magic. This function does the following:

    1. Prepare the parameters needed to call a kernel generated by `generate_kernel`.
       This includes both the launch configuration and the direct kernel function
       parameters.

    2. Call the kernel

    3. Ensure that the results are returned in the correct shape.

    This function does so many things at once because they all depend on the same
    "shape logic", which is harder to get right when split up all over the place.

    The reason behind this is that the CUDA kernels themselves only support inputs with
    equal dimensions for simplicity. This function makes the necessary calculations such
    that asymmetrical dimensions are flattened down to symmetrical ones before the
    kernel call, and restored later before returning the results.
    """

    input_rows, cam_rows = inputs.shape[-2], cam.shape[-2]
    columns = cam.shape[-1] // variant.cell_encoding_width

    # while the two innermost shapes of `inputs` and `cam` are `input_rows x columns`
    # and `cam_rows x columns` respectively, the remaining outer shapes hold all the
    # information on how the inputs are stacked and to be broadcasted.
    inputs_outer_shape, cam_outer_shape = inputs.shape[:-2], cam.shape[:-2]

    # sort the outer shapes by length. we expect the smaller shape (`sub_shape`) to be
    # a subset of the larger shape (`super_shape`) as the name suggests.
    # Specifically, the only way that `super_shape` can differ from `sub_shape` is by
    # extending it to the left, i.e. adding even higher dimensions. This behavior is
    # enforced within `check_cam_params`.
    #
    # For example:
    #
    # inputs_outer_shape = (4, 3), cam_outer_shape = (5, 3)
    # --> sub_shape = (), super_shape = ()
    #
    # inputs_outer_shape = (2, 4, 3), cam_outer_shape = (2, 4, 3)
    # --> sub_shape = (2), super_shape = (2)
    #
    # inputs_outer_shape = (5, 2, 4, 3), cam_outer_shape = (2, 4, 3)
    # --> sub_shape = (2), super_shape = (5, 2)
    sub_shape, super_shape = sorted([inputs_outer_shape, cam_outer_shape], key=len)

    # the "overhang" is the difference between `sub_shape` and a super shape, i.e.
    # the dimensions that remain to the left. If both outer shapes are equal and we have
    # symmetrical dimensions, the overhang will always be empty.
    #
    # Following the previous example:
    # inputs_outer_shape = (5, 2, 4, 3), cam_outer_shape = (2, 4, 3)
    # --> sub_shape = (2), super_shape = (5, 2)
    # --> overhang(inputs_outer_shape) = (5)
    # --> overhang(cam_outer_shape) = ()
    def overhang(shape: Tuple[int, ...]) -> Tuple[int, ...]:
        return shape[: -len(sub_shape)] if sub_shape else shape

    # these values tell us by which factor to extend `input_rows` or `cam_rows`
    # respectively when calling the kernel by reducing the overhanging dimensions
    # via multiplication.
    # at least one of these will always be equal to one because the outer shape is
    # equal to `sub_shape` and thereby the overhang is empty.
    inputs_overhang = prod(overhang(inputs_outer_shape))
    cam_overhang = prod(overhang(cam_outer_shape))

    is_reduction = op.is_reduction

    results_shape = super_shape + (
        (input_rows,) if is_reduction else (input_rows, cam_rows)
    )

    # allocate the required space for the results in the shape that they will
    # finally be returned as. Even though the kernel will always interpret `results`
    # in the context of an even stack, the buffer is the same size and the shape is
    # not relevant to the CUDA code.
    results = cp.zeros(
        results_shape,
        dtype=result_dtype,  # type: ignore (https://github.com/cupy/cupy/pull/7702)
    )

    # CuPy kindly provides this information such that we don't need to hardcode / guess
    # the value.
    # Depending on the GPU, this is typically 512 or 1024.
    max_threads_per_block = kernel.attributes["max_threads_per_block"]

    # a core = a single inputs/CAM pair within the operation stack
    # each core calculates a single point in a `input_rows x cam_rows` match matrix,
    # where we need to account for overhangs generates from asymmetrical stacks.
    threads_per_core = input_rows * inputs_overhang * cam_rows * cam_overhang

    # this is the edge case where we have less threads per core than what would fill
    # a single block.
    threads_per_block = min(threads_per_core, max_threads_per_block)

    # the `x` dimensions of `dim_block` and `dim_grid` generate `threads_per_core`
    # threads.
    # the `y` dimension of `dim_grid` accounts for the amount of cores, which given
    # an even stack is equal to `prod(sub_shape)`.
    dim_block = (threads_per_block, 1, 1)
    dim_grid = (ceil(threads_per_core / threads_per_block), prod(sub_shape), 1)

    kernel_args = (
        # make sure to move inputs and CAM to the GPU, this no-ops if they
        # are already on the device
        cp.asarray(inputs.ravel()),
        cp.asarray(cam.ravel()),
        columns,
        # the CAM kernels operate on even shapes, so the overhang needs to be
        # accounted for by extending the rows.
        input_rows * inputs_overhang,
        cam_rows * cam_overhang,
        results,
    )

    if is_reduction:
        assert reduction_values is not None
        kernel_args += (cp.asarray(reduction_values.ravel()),)

    # call the CUDA kernel. this will mutate `results` in place.
    kernel(dim_grid, dim_block, kernel_args)

    # this is the final special case where the kernel has placed the correct values
    # in `results` but it needs to be rearranged such that they also appear in the
    # correct place.
    # there are three different cases for the shapes of the parameters taken by the
    # CAM functions:
    #
    # 1. The shapes of `inputs` and `cam` are equal.
    # 2. The shape of `inputs` supersets the shape of `cam`.
    # 3. The shape of `cam` supersets the shape of `inputs`.
    #
    # case 1 works fine because the kernels themselves are built to directly handle it.
    # for case 2 and 3 each, `results` needs to be reshaped and transposed such that
    # the results correctly propagate across the overhanging dimensions.
    #
    # the algorithm here is as follows:
    #
    # 1. reshape `results` such that the overhanging dimensions are moved in front of
    #    the dimension that was initially modified during the kernel call because of
    #    the overhang. For case 2, this is in front of `input_rows` and for case 3,
    #    in front of `cam_rows`.
    #
    # 2. transpose `results` such that the dimensions go back into correct order.

    # case 1, nothing needs to be done
    if len(inputs_outer_shape) == len(cam_outer_shape):
        return results

    match_dims = len(results.shape)
    sub_shape_dims = len(sub_shape)

    # case 2
    if len(inputs_outer_shape) > len(cam_outer_shape):
        # move the overhang behind the sub shape and in front of the input rows
        results = results.reshape(
            (*sub_shape, *overhang(super_shape), input_rows, cam_rows)
        )
        # transpose the dimensions back into the right order
        results = results.transpose(
            *range(sub_shape_dims, match_dims - 2),  # move overhang in front
            *range(sub_shape_dims),  # move sub shape behing overhang
            match_dims - 2,  # keep input rows in place
            match_dims - 1,  # keep cam rows in place
        )

    # case 3
    else:
        # move the overhang behind the sub shape and the input rows and in front of the
        # cam rows
        results = results.reshape(
            (*sub_shape, input_rows, *overhang(super_shape), cam_rows)
        )
        # transpose the dimensions back into the right order
        results = results.transpose(
            *range(sub_shape_dims + 1, match_dims - 1),  # move overhang in front
            *range(sub_shape_dims),  # move sub shape behind overhang
            sub_shape_dims,  # move input rows back in front of cam rows
            match_dims - 1,  # keep cam rows in place
        )

    return results
